#include "mmc.h"

// Buffers to store commands and responses
int command_buf[9] ; 
int response_buf[9];

// Count to keep track of the interrupts
extern int interrupt_count ;

// Flag to indicate that 80 cycles needed for initialization are 
// complete (SPI DMA interrupt)
extern volatile int done_80_cycles ;

// Flag to indicate that data block is received using DMA
extern volatile int data_block_received ;

// Flag to indicate that data block is transmitted using DMA
extern volatile int data_block_transmitted ;

// Response value from the MMC Card.
int response;

// Flag used to indicate that token value (0xFE) is received.
extern int token_flag ;

// Flag used to indicate that data block is written to MMC successfully
extern volatile int data_response_flag ;

// Command that is being sent to MMC
volatile int command = 0;

// Function to Initialize MMC Card
// This includes initial 80 cycles, CMD0 and CMD1
void mmc_init()
{
	int i;
	
	// Initialize SPI for sending 80 or more dummy cycles.
	init_spi();

	// Interrupt dispatcher for SPI Core Rx ISR
	interrupts (SIG_SPIH, command_isr);

	/***************************************************************
	
				 Send Command0 - CMD_GO_IDLE_STATE
	
	****************************************************************/
	response = 0xff;
	// Continuously send the same command until the card sends a
	// valid response (i.e Card in IDLE State)
	while(response != R1_IN_IDLE_STATE)
	{
		// Initially CS is made logic "HI".
		SRU(HIGH, DAI_PB09_I);
	
		// Enable SPI in Core mode
		*pSPICTL = WL8 | MSBF | SPIMS | OPD | SMLS | GM | CPHASE | CLKPL ;

		mmc_send_cmd(CMD_GO_IDLE_STATE,0);
		response = mmc_get_response();
	}

	/***************************************************************
	
				 Send Command1 - CMD_SEND_OP_COND
	
	****************************************************************/
	response = 0xff;
	// Continuously send the same command until the card comes out
	// of idle state
	while(response != 0)
	{
		// Initially CS is made logic "HI".
		SRU(HIGH, DAI_PB09_I);
		
		// Enable SPI in Core mode
		*pSPICTL = WL8 | MSBF | SPIMS | OPD | SMLS | GM | CPHASE | CLKPL ;

		mmc_send_cmd(CMD_SEND_OP_COND,0);
		response = mmc_get_response();
	}
}

// Function to send dummy 80 cycles for initializing MMC Card
void init_spi()
{
	int i;
	int dummy[11];

	// Fill up dummy buffer with 0xff
	for(i=0;i<11;i++)
		dummy[i] = 0xff;
	
	// Initially CS is made logic "HI".
	SRU(HIGH, DAI_PB09_I);
	
	// Enable SPI DMA interrupt
	interrupts (SIG_SPIH, mmc_init_isr);
	
	// Clear SPI Control Registers
	*pSPICTL = 0x0;
	*pSPIDMAC = 0x0;
	
	// SPI Baud rate should be less than 400 KHz during the 
	// initialization process
	*pSPIBAUD = 800; // BAUDR = 400 => SPI Baud Rate = 200 / (4*400) = 125 KHz
	
	// Initialize DMA registers
	*pIISPI = (int)dummy;
	*pIMSPI = 1;
	*pCSPI = 11;

	// Program Control Registers
	*pSPICTL = SPIEN | SENDZ | SPIMS | MSBF | TIMOD2 | WL8 | CPHASE | SMLS | OPD | CLKPL; 	
	*pSPIDMAC = SPIDEN | INTEN ;
	
	// Wait for SPI DMA interrupt
	while(done_80_cycles ==0)
	{
	}
}

// Function to send commands to the MMC Card
void mmc_send_cmd(int cmd, int arg)
{
	command = cmd;
	// Filling up the command buffer
	command_buf[0] = 0x40 | cmd;
	command_buf[1] = (arg & 0xff000000) >> 24;
	command_buf[2] = (arg & 0xff0000) >> 16;
	command_buf[3] = (arg & 0xff00) >> 8;
	command_buf[4] = arg & 0xff;
	
	// CMD0 should be 0x40 00 00 00 00 95
	if(command == 0)
		command_buf[5] = 0x95;
	else
		command_buf[5] = 0x01;
		
	command_buf[6] = 0xff;
	command_buf[7] = 0xff;
	command_buf[8] = 0xff;
		
	// Enable SPI	
	*pSPICTL |= SPIEN ;

	// Assert Chip Select logic "LO" to select MMC Card.
	// Asserting Chip select logic "LO" during CMD0 will configure the
	// MMC card in SPI mode.
	SRU(LOW, DAI_PB09_I);

	// Load TX buffer with first byte
	*pTXSPI = command_buf[0];
}

// Function to return the response from the MMC Card
int mmc_get_response()
{
	// Wait till the MMC Card sends its response
	while(interrupt_count<9)
	{
	}
		
	// Clear the interrupt counter
	interrupt_count = 0 ;
	return response_buf[7];
}

// Function to set the block length
void mmc_set_block_length(int length)
{
	// Interrupt dispatcher for SPI Core Rx ISR
	interrupts (SIG_SPIH, command_isr);

	/***************************************************************
	
				 Send Command16 - CMD_SET_BLOCKLEN
	
	****************************************************************/
	response = 0xff;
	
	// Initially CS is made logic "HI".
	SRU(HIGH, DAI_PB09_I);
	
	// Enable SPI in Core mode
	*pSPICTL = WL8 | MSBF | SPIMS | OPD | SMLS | GM | CPHASE | CLKPL ;

	mmc_send_cmd(CMD_SET_BLOCKLEN,512);
	response = mmc_get_response();
	if(response!=0)
	{
		printf("Set Block Length command not successful\n");
		printf("Response from MMC: 0x%x",response);
		exit(1);
	}
}

// Function to read data from MMC Card
void mmc_read_data(int address, int length, int *destination)
{
	// Initially CS is made logic "HI".
	SRU(HIGH, DAI_PB09_I);
	
	// Interrupt dispatcher for SPI Core Rx ISR
	interrupts (SIG_SPIH, command_isr);

	/***************************************************************
	
				 Send Command17 - CMD_READ_BLOCK
	
	****************************************************************/
	response = 0xff;

	// Enable SPI in Core mode
	*pSPICTL = WL8 | MSBF | SPIMS | OPD | SMLS | GM | CPHASE | CLKPL ;

	mmc_send_cmd(CMD_READ_BLOCK,address);
	response = mmc_get_response();
	
	if(response!=0)
	{
		printf("Read Block command not successful\n");
		printf("Response from MMC: 0x%x",response);
		exit(1);
	}
	
	// Initially CS is made logic "HI".
	SRU(HIGH, DAI_PB09_I);
	
	// Enable SPI in Core mode
	// Wait till the DSP receives the token word
	interrupts (SIG_SPIH, wait_for_token);
	*pSPICTL = WL8 | MSBF | SPIMS | OPD | SMLS | GM | CPHASE | CLKPL | SPIEN;
	SRU(LOW, DAI_PB09_I);
	*pTXSPI = 0xff;
	while(token_flag == 0)
	{
	}
	
	SRU(HIGH, DAI_PB10_I);
	
	SRU(HIGH, DAI_PB09_I);
	
	// Clear SPI DMA Control Register
	*pSPIDMAC = 0x0;

	// Enable SPI Rx DMA interrupt for block reception
	interrupts (SIG_SPIH, read_block_isr);
	
	// Initialize DMA parameters
	*pIISPI = (int)destination;
	*pIMSPI = 1;
	*pCSPI = length + 2 ;

	SRU(LOW, DAI_PB09_I);
	
	// Program Control Registers
	*pSPICTL = SPIEN | SPIMS | MSBF | TIMOD2 | WL8 | CPHASE | SMLS | OPD | CLKPL | GM; 	
	*pSPIDMAC = SPIDEN | INTEN | SPIRCV ;
	
	// Wait for SPI DMA interrupt
	while(data_block_received ==0)
	{
	}
}

// Function to write data to MMC
void mmc_write_data(int address, int *source)
{
	// Initially CS is made logic "HI".
	SRU(HIGH, DAI_PB09_I);

	// Interrupt dispatcher for SPI Core Rx ISR
	interrupts (SIG_SPIH, command_isr);

	/***************************************************************
	
				 Send Command24 - CMD_WRITE_BLOCK
	
	****************************************************************/
	
	response = 0xff;

	// Enable SPI in Core mode
	*pSPICTL = WL8 | MSBF | SPIMS | OPD | SMLS | GM | CPHASE | CLKPL ;

	mmc_send_cmd(CMD_WRITE_BLOCK,address);
	response = mmc_get_response();
	
	if(response!=0)
	{
		printf("Write command not successful\n");
		printf("Response from MMC: 0x%x",response);
		exit(1);
	}
	
	// Initially CS is made logic "HI".
	SRU(HIGH, DAI_PB09_I);
	
	// Clear SPI DMA Control Register
	*pSPIDMAC = 0x0;
	
	// Enable SPI Tx DMA interrupt
	interrupts (SIG_SPIH, write_block_isr);
	
	SRU(LOW, DAI_PB09_I);

	// Initialize DMA parameters
	*pIISPI = (int)source;
	*pIMSPI = 1;
	*pCSPI = 515 ;

	// Program Control Registers
	*pSPICTL = SPIEN | SPIMS | MSBF | TIMOD2 | WL8 | CPHASE | SMLS | OPD | CLKPL | GM; 	
	*pSPIDMAC = SPIDEN | INTEN ;
	
	// Wait for SPI DMA interrupt
	while(data_block_transmitted ==0)
	{
	}	

	SRU(HIGH, DAI_PB09_I);

	// Wait till the data is written sucessfully into the 
	// MMC Card
	interrupts (SIG_SPIH, wait_for_data_response);
	*pSPICTL = WL8 | MSBF | SPIMS | OPD | SMLS | GM | CPHASE | CLKPL | SPIEN;
	SRU(LOW, DAI_PB09_I);
	*pTXSPI = 0xff;
	
	while(data_response_flag == 0)
	{
	}
}

// Function to turn the CRC check on or off
// Pass '1' for ON and '0' for OFF
void mmc_crc_on_off(int on_off)
{
	// Interrupt dispatcher for SPI Core Rx ISR
	interrupts (SIG_SPIH, command_isr);

	/***************************************************************
	
				 Send Command59 - CMD_CRC_ONOFF
	
	****************************************************************/
	response = 0xff;
	
	// Initially CS is made logic "HI".
	SRU(HIGH, DAI_PB09_I);
	
	// Enable SPI in Core mode
	*pSPICTL = WL8 | MSBF | SPIMS | OPD | SMLS | GM | CPHASE | CLKPL ;

	mmc_send_cmd(CMD_CRC_ONOFF,on_off);
	response = mmc_get_response();
	if(response!=0)
	{
		printf("Set CRC ON/OFF command not successful\n");
		printf("Response from MMC: 0x%x",response);
		exit(1);
	}
}
